/* ------------------------------------------------------------------------- *
 *
 * Implementation of Algorithms HITRO-box and HITRO-plate
 *
 * Roman Karawatzki, Josef Leydold, and Klaus Poetzelberger (2006).
 * Automatic Markov Chain Monte Carlo Procedures for Sampling from 
 * Multivariate Distributions.
 *
 * (c) 2006, Josef Leydold
 *
 * ------------------------------------------------------------------------- *
 *
 * ChangeLog:
 *
 * - Version 1.0 [Mon May 22 11:20:31] Josef Leydold:
 *      First release
 *
 * ------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* Load header files                                                         */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "hitro.h"

/* ------------------------------------------------------------------------- */
/* We use two types of coordinates:                                          */
/*                                                                           */
/* (1) xy coordinates:                                                       */
/*   argument (x_1,...,x_dim) is stored in double array '*x' of size 'dim',  */
/*   value of the PDF is stored in a double variable y.                      */
/* (2) vu coordinates (Ratio-of-uniforms coordinates):                       */
/*   coordinates stored in double array '*vu' of size 'dim'+1:               */
/*   (v,u_1,...,u_dim)                                                       */  
/* ------------------------------------------------------------------------- */

/* ------------------------------------------------------------------------- */
/* Constants                                                                 */

/* finding bounding rectangle adaptively */
#define BR_START       (1.e-3)  /* size of starting rectangle 
				   (must be positive!)                       */
#define BR_MULTIPLIER  (1.2)    /* multiplier for increasing rectangle
				   (must be greater than one)                */

/* ------------------------------------------------------------------------- */
/* structure for storing chain                                               */

/* prototype for PDF             */
typedef double PDF (int dim, double *x);   

/* HITRO chain */
struct hitro_chain {
  int use_coord;     /* 1 = use coordinate direction ("Gibbs") sampling      */
                     /* 0 = use random direction sampling                    */
  int dim;           /* dimension of distribution                            */
  double *state;     /* current state of chain (in vu coordinates)           */
  double *vu;        /* aux working array (next point in uv coordinates)     */
  double *x;         /* aux working array (next point in xy coordinates)     */
  double *direction; /* aux working array (random direction for next point)  */
  double *center;    /* center of distribution (in xy coordinates)           */
  double *vumax;     /* right upper bound bounding rectangle (in uv coord.)  */
  double *vumin;     /* right upper bound bounding rectangle (in uv coord.)  */
  double pdfmax;     /* upper bound for PDF                                  */
  double r;          /* parameter for Ratio-of-uniforms transformation       */
  int coord;         /* current coordinate for coordinate direction sampling */
  int adaptiverect;  /* 1 = use adaptive expanding of bounding box / plate   */
                     /*     [ "Slice Sampler" ]                              */
                     /* 0 = use given bounding rectangle / upper bound       */
  int adaptiveline;  /* 1 = use adaptive sampling from line segement         */
                     /* 0 = use simple sampling                              */
  int use_box;       /* 1 = use bounding rectangle ("bounding box")          */
                     /* 0 = use upper bound of PDF ("bounding plate")        */
  PDF *pdf;          /* pointer to PDF of distribution                       */
  UNIFORM *urng;     /* pointer to uniform RNG                               */
};

/* ------------------------------------------------------------------------- */
/* auiliary functions                                                        */

static void xy_to_vu( HITRO chain, const double *x, double y, double *vu );
/* transform xy coordinates into vu coordinates                              */

static void vu_to_x( HITRO chain, const double *vu, double *x );
/* transform vu coordinates into xy coordinates (y coordinate omiited)       */

static int vu_is_inside_region( HITRO chain, const double *vu );
/* check whether point is inside RoU acceptance region                       */

static void random_unitvector( HITRO chain, double *direction );
/* random vector uniformly distributed on the sphere                         */

static double random_normal( UNIFORM *urng );
/* standard normal random variate                                            */

/* ------------------------------------------------------------------------- */
/* Create and destroy HITRO chains                                           */
/* ------------------------------------------------------------------------- */

static HITRO 
hitro_create ( int dim, double *center, double *start,
	       double pdfmax, double *llbr, double *rubr, 
	       int adaptive, double r,
	       double pdf(int,double*), UNIFORM urng,
	       int use_box, int use_coord )
     /*----------------------------------------------------------------------*/
     /* create a new HITRO chain.                                            */
     /*                                                                      */
     /* parameters:                                                          */
     /*   dim       ... dimension of distribution                            */
     /*   center    ... center of distribution (mode or "typical point")     */
     /*                 NULL is interpreted as (0,0,...,0)                   */
     /*   start     ... starting point of chain. PDF(start) must not be 0    */
     /*                 if NULL, then 'center' is used as starting point     */
     /*   pdfmax    ... upper bound for PDF (non-positive = unknown)         */
     /*   llbr      ... left lower point of bounding rectangle               */
     /*   rubr      ... right upper point of bounding rectangle              */
     /*                 arrays of size 'dim'+1, storing ordering:            */
     /*                 v,u_1,u_2,...,u_dim                                  */
     /*                 Notice: bounding rectangle must be computed for      */
     /*                 PDF(x-center). vmin must be 0.                       */
     /*   adaptive  ... 1 = use adaptive sampling from line segement         */
     /*                 0 = use simple sampling                              */
     /*   r         ... parameter for Ratio-of-uniforms transformation       */
     /*   pdf       ... pointer to PDF of distribution                       */
     /*   urng      ... pointer to uniform RNG                               */
     /*   use_box   ... 1 = use bounding rectangle ("bounding box")          */
     /*                 0 = use upper bound of PDF ("bounding plate")        */
     /*   use_coord ... 1 = use coordinate direction ("Gibbs") sampling      */
     /*                 0 = use random direction sampling                    */
     /*----------------------------------------------------------------------*/
{
  HITRO chain;
  int i;
  double y;

  /* check arguments */
  if (dim < 1 || r <= 0.0) 
    return NULL;

  /* allocate memory for chain */
  chain = malloc( sizeof( struct hitro_chain ) );
  if (chain == NULL)
    return NULL;
  chain->state = malloc( (dim+1) * sizeof( double ) );
  chain->direction = malloc( (dim+1) * sizeof( double ) );
  chain->vu = malloc( (dim+1) * sizeof( double ) );
  chain->vumax = malloc( (dim+1) * sizeof( double ) );
  chain->vumin = malloc( (dim+1) * sizeof( double ) );
  chain->center = malloc( dim * sizeof( double ) );
  chain->x = malloc( dim * sizeof( double ) );

  /* store data */
  chain->dim = dim;
  chain->r = r;
  chain->pdf = pdf;
  chain->urng = urng;
  chain->use_box = use_box;
  chain->use_coord = use_coord;
  chain->adaptiveline = adaptive;
  chain->adaptiverect = 1;
  chain->coord = 0;
 
  /* center */
  if (center)
    memcpy( chain->center, center, dim*sizeof(double));
  else { /* use 0 vector */
    center = chain->center;
    for (i=0; i<dim; i++) center[i] = 0.0;
  }

  /* starting point */
  if (start == NULL) start = center;

  /* PDF at starting point */
  y = pdf(dim,start);
  if (y <= 0.0) {
    /* we need a proper starting point inside support of the PDF */
    hitro_delete( chain );
    return NULL;
  }

  /* transform starting point into uv coordinates and store as initial state */
  xy_to_vu( chain, start, y / 2.0, chain->state );
  memcpy( chain->vu, chain->state, (dim + 1)*sizeof(double) );

  /* upper bound for pdf */
  if (pdfmax > 0.0) {
    /* upper bound given */
    chain->adaptiverect = 0;
    chain->pdfmax = pdfmax;
  }
  else {
    /* no upper bound given --> use adaptive upper bound */
    chain->adaptiverect = 1;
    chain->pdfmax = y; /* starting point for searching upper bound */
    }
  /* store bounding rectangle */
  chain->vumin[0] = 0.0;
  chain->vumax[0] = pow(chain->pdfmax, 1.0/(r * dim + 1.0));

  /* bounding box or just upper bound */
  if (use_box) {
    /* bounding box */
    if (llbr != NULL && rubr != NULL) {
      chain->adaptiverect = 0;
      for (i=0; i<dim+1; i++) chain->vumin[i] = llbr[i];
      for (i=0; i<dim+1; i++) chain->vumax[i] = rubr[i];
    }
    else {
      chain->adaptiverect = 1;
      chain->vumin[0] = 0.0;
      for (i=1; i<dim+1; i++) chain->vumin[i] = -BR_START;
      for (i=1; i<dim+1; i++) chain->vumax[i] = BR_START;
    }
  }

  /* o.k. */
  return chain;

} /* end of hitro_create() */

/* ------------------------------------------------------------------------- */

int hitro_delete ( HITRO chain )
     /*----------------------------------------------------------------------*/
     /* delete HITRO chain.                                                  */
     /*----------------------------------------------------------------------*/
{
  if (chain == NULL)
    return 1;

  free( chain->state );
  free (chain->vu);
  free (chain->vumax);
  free (chain->vumin);
  free (chain->x);
  free( chain->center );
  free (chain->direction);
  free( chain );

  return 0;
} /* end of hitro_delete() */

/* ------------------------------------------------------------------------- */

HITRO hitro_plate_create ( int dim, double *center, double *start, double pdfmax, 
			   double pdf(int,double*), UNIFORM urng, double r )
     /*----------------------------------------------------------------------*/
     /* create a new HITRO chain for Algorithm HITRO-plate                   */
     /*                                                                      */
     /* parameters:                                                          */
     /*   dim       ... dimension of distribution                            */
     /*   center    ... center of distribution (mode or "typical point")     */
     /*                 NULL is interpreted as (0,0,...,0)                   */
     /*   start     ... starting point of chain. PDF(start) must not be 0    */
     /*                 if NULL, then 'center' is used as starting point     */
     /*   pdfmax    ... upper bound for PDF (non-positive = unknown)         */
     /*   pdf       ... pointer to PDF of distribution                       */
     /*   urng      ... pointer to uniform RNG                               */
     /*   r         ... parameter for Ratio-of-uniforms transformation       */
     /*----------------------------------------------------------------------*/
{
  /* call generic constructor */
  return hitro_create(
     dim,            /* dimension of distribution                            */
     center,         /* center of distribution (mode or "typical point")     */
     start,          /* starting point of chain                              */
     pdfmax,         /* upper bound for PDF                                  */
     NULL,           /* bounding rectangle not used (set to 'unknown')       */
     NULL,           /* bounding rectangle not used (set to 'unknown')       */
     1,              /* use adaptive sampling from line segement             */
     r,              /* parameter for Ratio-of-uniforms transformation       */
     pdf,            /* pointer to PDF of distribution                       */
     urng,           /* pointer to uniform RNG                               */
     0,              /* use upper bound of PDF ("bounding plate")            */
     0               /* use random direction sampling                        */
     );
} /* end of hitro_plate_create */

/* ------------------------------------------------------------------------- */

HITRO hitro_box_create ( int dim, double *center, double *start, double *llbr, double *rubr,
			   double pdf(int,double*), UNIFORM urng, double r, int adaptive, int use_coord )
     /*----------------------------------------------------------------------*/
     /* create a new HITRO chain for Algorithm HITRO-box                     */
     /*                                                                      */
     /* parameters:                                                          */
     /*   dim       ... dimension of distribution                            */
     /*   center    ... center of distribution (mode or "typical point")     */
     /*                 NULL is interpreted as (0,0,...,0)                   */
     /*   start     ... starting point of chain. PDF(start) must not be 0    */
     /*                 if NULL, then 'center' is used as starting point     */
     /*   llbr      ... left lower point of bounding rectangle               */
     /*   rubr      ... right upper point of bounding rectangle              */
     /*                 arrays of size 'dim'+1, storing ordering:            */
     /*                 v,u_1,u_2,...,u_dim                                  */
     /*                 Notice: bounding rectangle must be computed for      */
     /*                 PDF(x-center). vmin must be 0.                       */
     /*   pdf       ... pointer to PDF of distribution                       */
     /*   urng      ... pointer to uniform RNG                               */
     /*   r         ... parameter for Ratio-of-uniforms transformation       */
     /*   adaptive  ... 1 = use adaptive sampling from line segement         */
     /*                 0 = use simple sampling                              */
     /*   use_coord ... 1 = use coordinate direction ("Gibbs") sampling      */
     /*                 0 = use random direction sampling                    */
     /*----------------------------------------------------------------------*/
{
  /* call generic constructor */
  return hitro_create(
     dim,            /* dimension of distribution                            */
     center,         /* center of distribution (mode or "typical point")     */
     start,          /* starting point of chain                              */
     0.,             /* upper bound for PDF not used (set to 'unknown')      */ 
     llbr,           /* left lower point of bounding rectangle               */
     rubr,           /* right upper point of bounding rectangle              */
     adaptive,       /* do not / use adaptive sampling from line segement    */
     r,              /* parameter for Ratio-of-uniforms transformation       */
     pdf,            /* pointer to PDF of distribution                       */
     urng,           /* pointer to uniform RNG                               */
     1,              /* use bounding rectangle ("bounding box")              */
     use_coord       /* use coordinate direction / random direction sampling */
     );
} /* end of hitro_plate_create */

/* ------------------------------------------------------------------------- */
/* Run HITRO chains                                                          */
/* ------------------------------------------------------------------------- */

static int hitro_randomdir_sample( HITRO chain, double *x )
     /*----------------------------------------------------------------------*/
     /* sample from generator                                                */
     /*----------------------------------------------------------------------*/
{
#define new_point(ll)  { int j; for (j=0;j<chain->dim+1;j++) chain->vu[j] = chain->state[j]+(ll)*chain->direction[j]; }

  int i,k,d;
  double lambda, lb[2];    /* line segments, endpoints */
  double U;                /* uniform random number */

  /* entire bounding rectangle or just covering plate ? */
  d = (chain->use_box) ? chain->dim+1 : 1;

  /* new random direction */
  random_unitvector( chain, chain->direction );
    
  /* --- l.h.s. and r.h.s. endpoint of line segment --- */
  lb[0] = -DBL_MAX;
  lb[1] = DBL_MAX;
  for (i=0; i<d; i++) {
    if (chain->direction[i] != 0.0) {
      lambda = (chain->vumin[i] - chain->state[i]) / chain->direction[i];
      if (lambda>0 && lambda<lb[1]) lb[1] = lambda;
      if (lambda<0 && lambda>lb[0]) lb[0] = lambda;
      lambda = (chain->vumax[i] - chain->state[i]) / chain->direction[i];
      if (lambda>0 && lambda<lb[1]) lb[1] = lambda;
      if (lambda<0 && lambda>lb[0]) lb[0] = lambda;
    }
  }

  /* --- adaptive bounding rectangle --- */
  if (chain->adaptiverect) {
    /* check whether endpoint is outside of region */
    for (k=0; k<2; k++) {
      int update = 0;
        while (1) {
          new_point(lb[k]);
          if (! vu_is_inside_region(chain,chain->vu) )
            break;
          update = 1;
	  lb[k] *= BR_MULTIPLIER;
	}
        if (update) {
          /* we have to update vumin and vmax */
          new_point(lb[k]);
          for (i=0; i<d; i++) {
            if (chain->vu[i] < chain->vumin[i] && i!=0) chain->vumin[i] =  chain->vu[i];
            if (chain->vu[i] > chain->vumax[i])         chain->vumax[i] =  chain->vu[i];
            /* remark: vmin is always 0 */
          }
        }
      }
    }

  /* --- new coordinate for next point --- */
  while (1) {
    U = uniform(chain->urng);
    lambda = U * lb[0] + (1.-U) * lb[1];
    new_point(lambda);
    if (vu_is_inside_region(chain,chain->vu) )
      break;

    /* --- adaptive line sampling --- */
    if (chain->adaptiveline) {
      if (lambda < 0) lb[0] = lambda;
      else            lb[1] = lambda;
    }
  }

  /* store new state */
  memcpy( chain->state, chain->vu, (chain->dim+1)*sizeof(double) );

  /* transform current state into point in original scale */
  vu_to_x( chain, chain->state, x );
  
  return 0;
  
#undef new_point
} /* end of hitro_randomdir_sample() */

/* ------------------------------------------------------------------------- */

static int hitro_coord_sample( HITRO chain, double *x )
     /*----------------------------------------------------------------------*/
     /* sample from generator                                                */
     /*----------------------------------------------------------------------*/
{

  double lmin, lmax, lmid;   /* l.h.s., r.h.s. endpoint and midpoint of line segment */
  double *vu = chain->vu;    /* pointer to auxiliary working array */
  int coord;                 /* direction */
  double U;                  /* uniform random number */

  /* update coordinate direction */
  coord = chain->coord = (chain->coord + 1) % (chain->dim + 1);

  /* l.h.s. and r.h.s. endpoint of line segment */
  lmin = chain->vumin[coord];
  lmax = chain->vumax[coord];

  /* --- adaptive bounding rectangle --- */
  if (chain->adaptiverect) {
    lmid = 0.5 * (lmin + lmax);
    /* check whether endpoint is outside of region */
    vu[coord] = lmax;
    while ( vu_is_inside_region(chain,vu) ) {
      lmax = lmid + (lmax-lmid) * BR_MULTIPLIER;
      chain->vumax[coord] = vu[coord] = lmax;
    }
    vu[coord] = lmin;
    /* no update of lmin in case of coord == 0 since v-coordinate always > 0 */
    while ( coord!=0 && vu_is_inside_region(chain,vu) ) {
      lmin = lmid + (lmin-lmid) * BR_MULTIPLIER;
      chain->vumin[coord] = vu[coord] = lmin;
    }
  }

  /* --- new coordinate for next point --- */
  while (1) {
    U = uniform(chain->urng);
    vu[coord] = U * lmin + (1.-U) * lmax;
    if (vu_is_inside_region(chain,vu) )
      break;

    /* --- adaptive line sampling --- */
    if (chain->adaptiveline) {
      if (chain->state[coord] < vu[coord])
	lmax = vu[coord];
      else
	lmin = vu[coord];
    }
  }

  /* store new state */
  chain->state[coord] = vu[coord];
  
  /* transform current state into point in original scale */
  vu_to_x( chain, chain->state, x );
  
  return 0;

} /* end of hitro_coord_sample() */

/* ------------------------------------------------------------------------- */

int hitro_sample( HITRO chain, double *x )
     /*----------------------------------------------------------------------*/
     /* sample from generator                                                */
     /*----------------------------------------------------------------------*/
{
  return (chain->use_coord ? 
	  hitro_coord_sample(chain,x) :      /* coordinate direction sampler */
	  hitro_randomdir_sample(chain,x)    /* random direction sampler     */
	  );
} /* end of hitro_sample() */

/* ------------------------------------------------------------------------- */
/* Auxiliary routines                                                        */
/* ------------------------------------------------------------------------- */

void xy_to_vu( HITRO chain, const double *x, double y, double *vu )
     /*----------------------------------------------------------------------*/
     /* transform point with coordinates xy in the original scale into       */
     /* point in RoU-scale with coordinates uv.                              */
     /*----------------------------------------------------------------------*/
{
  int i;
  int dim = chain->dim;
  double r = chain->r;
  const double *center = chain->center;
  double *u = vu+1;
  double v = vu[0]= pow(y, 1.0/(r * dim + 1.0));
  double vr = (r==1.0) ? v : pow(v,r);

  for (i=0; i<dim; i++)  u[i] = (x[i] - center[i]) * vr;

} /* end of xy_to_vu() */

/*---------------------------------------------------------------------------*/

void vu_to_x( HITRO chain, const double *vu, double *x )
     /*----------------------------------------------------------------------*/
     /* transform point with coordinates vu in RoU-scale into original       */
     /* scale with coordinates xy. The y coordinate is not of interest here  */
     /* and thus omitted.                                                    */
     /*----------------------------------------------------------------------*/
{
  int i;
  int dim = chain->dim;
  double r = chain->r;
  const double *center = chain->center;
  const double *u = vu+1;
  double v = vu[0];
  double vr;

  if (v<=0.) {
    /* we are outside of the domain --> return 0 */
    for (i=0; i<dim; i++)  x[i] = 0.0;
    return;
  }

  vr = (r==1.0) ? v : pow(v,r);
    
  for (i=0; i<dim; i++)  x[i] = u[i]/vr + center[i];

} /* end of vu_to_x() */

/*---------------------------------------------------------------------------*/

int vu_is_inside_region( HITRO chain, const double *vu )
     /*----------------------------------------------------------------------*/
     /* check whether point is inside RoU acceptance region                  */
     /*----------------------------------------------------------------------*/
{
  double y;

  /* v-coordinate */
  double v = vu[0];

  /* transform into original scale */
  vu_to_x( chain, vu, chain->x );

  /* PDF at x */
  y = (chain->pdf)(chain->dim,chain->x);

  /* we are outside of domain if x is not in support of PDF or */
  /* v is non-positive.                                        */
  if (y <= 0.0 || v <= 0.0) return 0;

  /* now check whether point is in region below PDF */
  return ( (v < pow(y,1.0/(chain->r * chain->dim + 1.0))) ? 1 : 0 );

} /* end of vu_is_inside_region() */

/*---------------------------------------------------------------------------*/

void random_unitvector( HITRO chain, double *direction )
     /*----------------------------------------------------------------------*/
     /* random vector uniformly distributed on the sphere                    */
     /*----------------------------------------------------------------------*/
{
  int i;
  int dim = chain->dim;
  double norm = 0.0;

  do {
    for (i=0; i<dim+1; i++) {
      direction[i] = random_normal(chain->urng);
      norm += direction[i] * direction[i];
    }
  } while (norm == 0.0);
    
  /* normalize direction vector */
  norm = sqrt(norm);
  for (i=0; i<dim+1; i++)
    direction[i] /= norm;

} /* end of random_unitvector() */

/*---------------------------------------------------------------------------*/

double random_normal ( UNIFORM *urng )
     /*----------------------------------------------------------------------*/
     /* standard normal random variate                                       */
     /* Method: acceptance-complement ratio                                  */
     /* Reference: W. Hoermann and G. Derflinger (1990).                     */
     /*            The ACR Methodfor generating normal random variables,     */
     /*            OR Spektrum 12 (1990), 181-185.                           */
     /*----------------------------------------------------------------------*/
{
#define c1    1.448242853
#define c2    3.307147487
#define c3    1.46754004
#define d1    1.036467755
#define d2    5.295844968
#define d3    3.631288474
#define hm    0.483941449
#define zm    0.107981933
#define hp    4.132731354
#define zp   18.52161694
#define phln  0.4515827053
#define hm1   0.516058551
#define hp1   3.132731354
#define hzm   0.375959516
#define hzmp  0.591923442

#define as    0.8853395638
#define bs    0.2452635696
#define cs    0.2770276848
#define b     0.5029324303
#define x0    0.4571828819
#define ym    0.187308492 
#define s     0.7270572718 
#define t     0.03895759111

  double X;
  double rn,x,y,z;

  do {
    y = uniform(urng);

    if (y>hm1) {
      X = hp*y-hp1; break; }
  
    else if (y<zm) {  
      rn = zp*y-1.0;
      X = (rn>0) ? (1.0+rn) : (-1.0+rn);
      break;
    } 

    else if (y<hm) {  
      rn = uniform(urng);
      rn = rn-1.0+rn;
      z = (rn>0) ? 2.0-rn : -2.0-rn;
      if ((c1-y)*(c3+fabs(z))<c2) {
        X = z; break; }
      else {  
        x = rn*rn;
        if ((y+d1)*(d3+x)<d2) {
          X = rn; break; }
        else if (hzmp-y<exp(-(z*z+phln)/2.0)) {
          X = z; break; }
        else if (y+hzm<exp(-(x+phln)/2.0)) {
          X = rn; break; }
      }
    }

    while (1) {
      x = uniform(urng);
      y = ym * uniform(urng);
      z = x0 - s*x - y;
      if (z>0) 
        rn = 2.0+y/x;
      else {
        x = 1.0-x;
        y = ym-y;
        rn = -(2.0+y/x);
      }
      if ((y-as+x)*(cs+x)+bs<0.0) {
        X = rn; break; }
      else if (y<x+t)
        if (rn*rn<4.0*(b-log(x))) {
          X = rn; break; }
    }
  } while(0);

  return X;

} /* end of random_normal() */

/* ------------------------------------------------------------------------- */

